@@ -331,7 +331,9 @@ remove_scan_regions (GtkSourceCompletionWordsBuffer *buffer,
 	GList *item;
 	gint span = end - start + 1;
 	
-	for (item = buffer->priv->scan_regions; item; item = g_list_next (item))
+	item = buffer->priv->scan_regions;
+
+	while (item != NULL)
 	{
 		ScanRegion *region = REGION_FROM_LIST (item);
 		
@@ -343,17 +345,25 @@ remove_scan_regions (GtkSourceCompletionWordsBuffer *buffer,
 				GList *remove = item;
 				scan_region_free (region);
 
-				item = g_list_previous (item);
+				item = g_list_next (item);
 			
 				/* Remove whole thing */
 				buffer->priv->scan_regions = 
 					g_list_delete_link (buffer->priv->scan_regions,
 						            remove);
+				continue;
 			}
 			else if (region->start <= end)
 			{
 				/* Top part of region in removed region */
-				region->start = end;
+				region->start = end + 1;
+				region->end -= span;
+			}
+			else
+			{
+				/* Fully decrease */
+				region->start -= span;
+				region->end -= span;
 			}
 		}
 		else if (region->end <= end && region->end > start)
@@ -365,6 +375,8 @@ remove_scan_regions (GtkSourceCompletionWordsBuffer *buffer,
 		{
 			region->end -= span;
 		}
+
+		item = g_list_next (item);
 	}
 }
 
@@ -456,7 +468,7 @@ add_scan_region (GtkSourceCompletionWordsBuffer *buffer,
 		/* Simply prepend, there was no overlap */
 		buffer->priv->scan_regions = 
 			g_list_insert_before (buffer->priv->scan_regions,
-			                      g_list_next (insert_after),
+			                      insert_after ? g_list_next (insert_after) : buffer->priv->scan_regions,
 			                      scan_region_new (start, end));
 	}
 	else
@@ -485,8 +497,10 @@ add_scan_region (GtkSourceCompletionWordsBuffer *buffer,
 	    !gtk_source_completion_words_library_is_locked (buffer->priv->library))
 	{
 		buffer->priv->idle_scan_id = 
-			g_idle_add ((GSourceFunc)idle_scan_regions,
-			            buffer);
+			g_idle_add_full (G_PRIORITY_LOW,
+			                 (GSourceFunc)idle_scan_regions,
+			                 buffer,
+			                 NULL);
 	}
 }
 
@@ -497,33 +511,44 @@ handle_text_inserted (GtkSourceCompletionWordsBuffer *buffer,
 {
 	gint pos = start;
 	GList *ptr = NULL;
+	GList *newlines = NULL;
+	GList *last = NULL;
 	
+	while (pos < end)
+	{
+		newlines = g_list_prepend (newlines, NULL);
+
+		if (last == NULL)
+		{
+			last = newlines;
+		}
+
+		++pos;
+	}
+
 	if (start > end)
 	{
 		ptr = g_list_nth (buffer->priv->lines,
 		                  start + 1);
 	}
-	
-	while (pos < end)
+
+	if (ptr != NULL)
 	{
-		/* Insert new empty lines */
-		if (ptr == NULL)
+		if (ptr->prev)
 		{
-			buffer->priv->lines = 
-				g_list_append (buffer->priv->lines,
-				               NULL);
-		}
-		else
-		{
-			buffer->priv->lines = 
-				g_list_insert_before (buffer->priv->lines,
-					              ptr,
-					              NULL);
+			ptr->prev->next = newlines;
+			newlines->prev = ptr->prev;
 		}
 
-		++pos;
+		newlines->next = ptr;
+		ptr->prev = last;
 	}
-	
+	else
+	{
+		buffer->priv->lines = g_list_concat (buffer->priv->lines,
+		                                     newlines);
+	}
+
 	/* Invalidate new region */
 	add_scan_region (buffer, 
 	                 start,
@@ -617,8 +642,10 @@ on_library_unlock (GtkSourceCompletionWordsBuffer *buffer)
 	    buffer->priv->scan_regions != NULL)
 	{
 		buffer->priv->idle_scan_id = 
-			g_idle_add ((GSourceFunc)idle_scan_regions,
-			            buffer);
+			g_idle_add_full (G_PRIORITY_LOW,
+			                 (GSourceFunc)idle_scan_regions,
+			                 buffer,
+			                 NULL);
 	}
 }
 
